// (c) 2002 by Jrn Loviscach

#include <math.h>

#include "bezier.h"
#include "adapt.h"
#include <assert.h>

Bezier::Bezier()
#ifndef NDEBUG
: hasData(false)
#endif
{}

Bezier::Bezier(const double* pX, const double* pY, const double* pInvD)
#ifndef NDEBUG
: hasData(true)
#endif
{
	int i;
	for(i=0; i<4; i++)
	{
		x[i] = pX[i];
		y[i] = pY[i];
		invD[i] = pInvD[i];
	}
}

Bezier::Bezier(const Bezier& bez)
#ifndef NDEBUG
: hasData(true)
#endif
{
	assert(bez.hasData);
	int i;
	for(i=0; i<4; i++)
	{
		x[i] = bez.x[i];
		y[i] = bez.y[i];
		invD[i] = bez.invD[i];
	}
}

Bezier::~Bezier()
{
}

double Bezier::GetX(int num) const
{
	assert(hasData);
	assert((num>=0)&&(num<=3));
	return x[num];
}

double Bezier::GetY(int num) const
{
	assert(hasData);
	assert((num>=0)&&(num<=3));
	return y[num];
}

double Bezier::GetInvD(int num) const
{
	assert(hasData);
	assert((num>=0)&&(num<=3));
	return invD[num];
}

double Bezier::FitJoin(const Bezier& b1, const Bezier& b2, double t)
{
	#ifndef NDEBUG
	hasData = true;
	#endif

	const double n = 2.0; // weight of shared (=middle) vertex

	// old vertices
	double a1x = b1.GetX(0);
	double b1x = b1.GetX(1);
	double c1x = b1.GetX(2);
	double d1x = b1.GetX(3);
	double a2x = b2.GetX(0);
	double b2x = b2.GetX(1);
	double c2x = b2.GetX(2);
	double d2x = b2.GetX(3);
	double a1y = b1.GetY(0);
	double b1y = b1.GetY(1);
	double c1y = b1.GetY(2);
	double d1y = b1.GetY(3);
	double a2y = b2.GetY(0);
	double b2y = b2.GetY(1);
	double c2y = b2.GetY(2);
	double d2y = b2.GetY(3);

	// new vertices, called a3x etc. in MuPAD calculation
	x[0] = a1x;
	x[3] = d2x;
	y[0] = a1y;
	y[3] = d2y;

	double t99 = t*t ;
	double t100 = t99*t99 ;
	double t101 = t100*t100 ;
    double t102 = t*6.0 ;
    double t103 = t99*20.0 ;
    double t104 = 38.0*t*t99 ;
    double t105 = t100*53.0 ;
    double t106 = 66.0*t*t100 ;
    double t107 = 64.0*t99*t100 ;
    double t108 = 36.0*t*t99*t100 ;
    double t109 = t101*9.0 ;
   	double t110 = 9.0*n*t99 ;
   	double t111 = 54.0*n*t*t99 ;
   	double t112 = 144.0*n*t100 ;
   	double t113 = 234.0*n*t*t100 ;
   	double t114 = 288.0*n*t99*t100 ;
    double t115 = 288.0*n*t*t99*t100 ;
    double t116 = 207.0*n*t101 ;
    double t117 = 90.0*n*t*t101 ;
    double t118 = 18.0*n*t99*t101 ;
    double t119 = t110 - t102 - t111 + t103 + t112 - t104 - t113 + t105 + t114 - t106 - t115 + t107 + t116 - t108 - t117 + t109 + t118 + 1.0 ;
    double t120 = 2.0*t99*b1x ;
    double t121 = 12.0*t*t99*c1x ;
    double t122 = 8.0*t100*b1x ;
    double t123 = 9.0*n*t*t99*t100*b1x ;
    double t124 = 9.0*n*t*t99*t100*c2x ;
   	double t125 = 3.0*n*t*t99*t100*d1x ;
  	double t126 = 9.0*n*t101*b2x ;
    double t127 = 9.0*n*t101*c1x ;
    x[1] = (-t120 + t121 - t122 + t123 - t124 + b2x + t125 - t126 - t127 - 3.0*t*a1x + t*b1x - 4.0*t*b2x + 2.0*t*c1x - 2.0*t*c2x - 3.0*n*t*a1x + 3.0*n*t*d1x + 13.0*t99*a1x + 6.0*t99*b2x - 6.0*t99*c1x + 8.0*t99*c2x + t99*d2x + 63.0*t100*a1x + 6.0*t101*a1x - 2.0*t100*b2x - 20.0*t100*c1x + 12.0*t100*c2x + 8.0*t100*d2x + 3.0*t101*d2x + 21.0*n*t99*a1x - 12.0*n*t99*d1x - 33.0*t*t99*a1x + 5.0*t*t99*b1x - 4.0*t*t99*b2x - 14.0*t*t99*c2x - 4.0*t*t99*d2x + 147.0*n*t100*a1x + 198.0*n*t101*a1x - 9.0*n*t100*b2x + 36.0*n*t100*c2x - 36.0*n*t100*d1x + 6.0*n*t100*d2x + 27.0*n*t101*d2x - 83.0*t*t100*a1x + 5.0*t*t100*b1x + 6.0*t*t100*b2x + 18.0*t*t100*c1x - 4.0*t*t100*c2x - 8.0*t*t100*d2x - 69.0*n*t*t99*a1x - 9.0*n*t*t99*c2x + 24.0*n*t*t99*d1x - 246.0*n*t*t100*a1x + 9.0*n*t*t100*b1x - 72.0*n*t*t101*a1x + 36.0*n*t*t100*b2x + 9.0*n*t*t100*c1x - 54.0*n*t*t100*c2x + 36.0*n*t*t100*d1x + 67.0*t99*t100*a1x - 24.0*n*t*t100*d2x - 18.0*n*t*t101*d2x - 3.0*t99*t100*b2x - 6.0*t99*t100*c1x + 6.0*t99*t100*d2x + 330.0*n*t99*t100*a1x - 18.0*n*t99*t100*b1x + 12.0*n*t99*t101*a1x - 54.0*n*t99*t100*b2x - 27.0*n*t99*t100*c1x + 36.0*n*t99*t100*c2x - 18.0*n*t99*t100*d1x + 39.0*n*t99*t100*d2x + 6.0*n*t99*t101*d2x - 30.0*t*t99*t100*a1x - 6.0*t*t99*t100*d2x - 318.0*n*t*t99*t100*a1x + 36.0*n*t*t99*t100*b2x + 27.0*n*t*t99*t100*c1x - 36.0*n*t*t99*t100*d2x)/t119 ;
	double t128 = 2.0*t99*b1y ;
	double t129 = 12.0*t*t99*c1y ;
    double t130 = 8.0*t100*b1y ;
    double t131 = 9.0*n*t*t99*t100*b1y ;
    double t132 = 9.0*n*t*t99*t100*c2y ;
    double t133 = 3.0*n*t*t99*t100*d1y ;
    double t134 = 9.0*n*t101*b2y ;
    double t135 = 9.0*n*t101*c1y ;
	y[1] = (-t130 + t131 - t132 + t133 - t134 + b2y - t135 - t128 + t129 - 3.0*t*a1y + t*b1y - 4.0*t*b2y + 2.0*t*c1y - 2.0*t*c2y - 3.0*n*t*a1y + 3.0*n*t*d1y + 13.0*t99*a1y + 6.0*t99*b2y - 6.0*t99*c1y + 8.0*t99*c2y + t99*d2y + 63.0*t100*a1y + 6.0*t101*a1y - 2.0*t100*b2y - 20.0*t100*c1y + 12.0*t100*c2y + 8.0*t100*d2y + 3.0*t101*d2y + 21.0*n*t99*a1y - 12.0*n*t99*d1y - 33.0*t*t99*a1y + 5.0*t*t99*b1y - 4.0*t*t99*b2y - 14.0*t*t99*c2y - 4.0*t*t99*d2y + 147.0*n*t100*a1y + 198.0*n*t101*a1y - 9.0*n*t100*b2y + 36.0*n*t100*c2y - 36.0*n*t100*d1y + 6.0*n*t100*d2y + 27.0*n*t101*d2y - 83.0*t*t100*a1y + 5.0*t*t100*b1y + 6.0*t*t100*b2y + 18.0*t*t100*c1y - 4.0*t*t100*c2y - 8.0*t*t100*d2y - 69.0*n*t*t99*a1y - 9.0*n*t*t99*c2y + 24.0*n*t*t99*d1y - 246.0*n*t*t100*a1y + 9.0*n*t*t100*b1y - 72.0*n*t*t101*a1y + 36.0*n*t*t100*b2y + 9.0*n*t*t100*c1y - 54.0*n*t*t100*c2y + 36.0*n*t*t100*d1y + 67.0*t99*t100*a1y - 24.0*n*t*t100*d2y - 18.0*n*t*t101*d2y - 3.0*t99*t100*b2y - 6.0*t99*t100*c1y + 6.0*t99*t100*d2y + 330.0*n*t99*t100*a1y - 18.0*n*t99*t100*b1y + 12.0*n*t99*t101*a1y - 54.0*n*t99*t100*b2y - 27.0*n*t99*t100*c1y + 36.0*n*t99*t100*c2y - 18.0*n*t99*t100*d1y + 39.0*n*t99*t100*d2y + 6.0*n*t99*t101*d2y - 30.0*t*t99*t100*a1y - 6.0*t*t99*t100*d2y - 318.0*n*t*t99*t100*a1y + 36.0*n*t*t99*t100*b2y + 27.0*n*t*t99*t100*c1y - 36.0*n*t*t99*t100*d2y)/t119 ;
	x[2] = (-t120 + t121 - t122 + t123 - t124 - t125 + c2x - t126 - t127 - 5.0*t*c2x - t*d2x + 5.0*t99*a1x - 3.0*t99*c1x + 15.0*t99*c2x + 5.0*t99*d2x + 58.0*t100*a1x + 3.0*t101*a1x - 20.0*t100*b2x - 17.0*t100*c1x + 17.0*t100*c2x + 23.0*t100*d2x + 6.0*t101*d2x + 3.0*n*t99*a1x + 9.0*n*t99*c2x - 3.0*n*t99*d1x - 26.0*t*t99*a1x + 6.0*t*t99*b1x + 8.0*t*t99*b2x - 23.0*t*t99*c2x - 15.0*t*t99*d2x + 93.0*n*t100*a1x - 9.0*n*t100*b1x + 135.0*n*t101*a1x - 45.0*n*t100*b2x - 9.0*n*t100*c1x + 90.0*n*t100*c2x - 21.0*n*t100*d1x + 45.0*n*t100*d2x + 90.0*n*t101*d2x - 70.0*t*t100*a1x + 4.0*t*t100*b1x + 18.0*t*t100*b2x + 12.0*t*t100*c1x - 5.0*t*t100*c2x - 25.0*t*t100*d2x - 24.0*n*t*t99*a1x + 9.0*n*t*t99*b2x - 45.0*n*t*t99*c2x + 15.0*n*t*t99*d1x - 9.0*n*t*t99*d2x - 210.0*n*t*t100*a1x + 27.0*n*t*t100*b1x - 42.0*n*t*t101*a1x + 90.0*n*t*t100*b2x + 36.0*n*t*t100*c1x - 90.0*n*t*t100*c2x + 9.0*n*t*t100*d1x + 48.0*t99*t100*a1x - 96.0*n*t*t100*d2x - 48.0*n*t*t101*d2x - 6.0*t99*t100*b2x - 3.0*t99*t100*c1x + 25.0*t99*t100*d2x + 291.0*n*t99*t100*a1x - 27.0*n*t99*t100*b1x + 6.0*n*t99*t101*a1x - 90.0*n*t99*t100*b2x - 54.0*n*t99*t100*c1x + 45.0*n*t99*t100*c2x + 3.0*n*t99*t100*d1x + 120.0*n*t99*t100*d2x + 12.0*n*t99*t101*d2x - 18.0*t*t99*t100*a1x - 18.0*t*t99*t100*d2x - 252.0*n*t*t99*t100*a1x + 45.0*n*t*t99*t100*b2x + 36.0*n*t*t99*t100*c1x - 114.0*n*t*t99*t100*d2x)/t119 ;
	y[2] = (-t130 + t131 - t132 - t133 - t134 - t135 + c2y - t128 + t129 - 5.0*t*c2y - t*d2y + 5.0*t99*a1y - 3.0*t99*c1y + 15.0*t99*c2y + 5.0*t99*d2y + 58.0*t100*a1y + 3.0*t101*a1y - 20.0*t100*b2y - 17.0*t100*c1y + 17.0*t100*c2y + 23.0*t100*d2y + 6.0*t101*d2y + 3.0*n*t99*a1y + 9.0*n*t99*c2y - 3.0*n*t99*d1y - 26.0*t*t99*a1y + 6.0*t*t99*b1y + 8.0*t*t99*b2y - 23.0*t*t99*c2y - 15.0*t*t99*d2y + 93.0*n*t100*a1y - 9.0*n*t100*b1y + 135.0*n*t101*a1y - 45.0*n*t100*b2y - 9.0*n*t100*c1y + 90.0*n*t100*c2y - 21.0*n*t100*d1y + 45.0*n*t100*d2y + 90.0*n*t101*d2y - 70.0*t*t100*a1y + 4.0*t*t100*b1y + 18.0*t*t100*b2y + 12.0*t*t100*c1y - 5.0*t*t100*c2y - 25.0*t*t100*d2y - 24.0*n*t*t99*a1y + 9.0*n*t*t99*b2y - 45.0*n*t*t99*c2y + 15.0*n*t*t99*d1y - 9.0*n*t*t99*d2y - 210.0*n*t*t100*a1y + 27.0*n*t*t100*b1y - 42.0*n*t*t101*a1y + 90.0*n*t*t100*b2y + 36.0*n*t*t100*c1y - 90.0*n*t*t100*c2y + 9.0*n*t*t100*d1y + 48.0*t99*t100*a1y - 96.0*n*t*t100*d2y - 48.0*n*t*t101*d2y - 6.0*t99*t100*b2y - 3.0*t99*t100*c1y + 25.0*t99*t100*d2y + 291.0*n*t99*t100*a1y - 27.0*n*t99*t100*b1y + 6.0*n*t99*t101*a1y - 90.0*n*t99*t100*b2y - 54.0*n*t99*t100*c1y + 45.0*n*t99*t100*c2y + 3.0*n*t99*t100*d1y + 120.0*n*t99*t100*d2y + 12.0*n*t99*t101*d2y - 18.0*t*t99*t100*a1y - 18.0*t*t99*t100*d2y - 252.0*n*t*t99*t100*a1y + 45.0*n*t*t99*t100*b2y + 36.0*n*t*t99*t100*c1y - 114.0*n*t*t99*t100*d2y)/t119 ;

	double b3x = x[1]; 
	double b3y = y[1]; 
	double c3x = x[2]; 
	double c3y = y[2]; 
	
	assert(fabs(x[0])<100000);
	assert(fabs(x[1])<100000);
	assert(fabs(x[2])<100000);
	assert(fabs(x[3])<100000);
	assert(fabs(y[0])<100000);
	assert(fabs(y[1])<100000);
	assert(fabs(y[2])<100000);
	assert(fabs(y[3])<100000);


	double a1d = b1.GetInvD(0);
	double b1d = b1.GetInvD(1);
	double c1d = b1.GetInvD(2);
	double d1d = b1.GetInvD(3);
	double a2d = b2.GetInvD(0);
	double b2d = b2.GetInvD(1);
	double c2d = b2.GetInvD(2);
	double d2d = b2.GetInvD(3);
	// possible improvement: optimize invD spline like x and y
	invD[0] = 0.7*a1d + 0.3*b1d;
	invD[1] = 0.3*c1d + 0.7*d1d;
	invD[2] = 0.7*a2d + 0.3*b2d;
	invD[3] = 0.3*c2d + 0.7*d2d;

	return pow(b1x - t*b3x - a1x*(-t + 1.0), 2.0) + pow(b1y - t*b3y - a1y*(-t + 1.0), 2.0) + pow(c2x - t*d2x - c3x*(-t + 1.0), 2.0) + pow(c2y - t*d2y - c3y*(-t + 1.0), 2.0) + pow(c1x - t*(t*c3x + b3x*(-t + 1.0)) - (-t + 1.0)*(t*b3x + a1x*(-t + 1.0)), 2.0) + pow(b2x - t*(t*d2x + c3x*(-t + 1.0)) - (-t + 1.0)*(t*c3x + b3x*(-t + 1.0)), 2.0) + pow(c1y - t*(t*c3y + b3y*(-t + 1.0)) - (-t + 1.0)*(t*b3y + a1y*(-t + 1.0)), 2.0) + pow(b2y - t*(t*d2y + c3y*(-t + 1.0)) - (-t + 1.0)*(t*c3y + b3y*(-t + 1.0)), 2.0) + n*pow(d1x - t*(t*(t*d2x + c3x*(-t + 1.0)) + (-t + 1.0)*(t*c3x + b3x*(-t + 1.0))) - (-t + 1.0)*(t*(t*c3x + b3x*(-t + 1.0)) + (-t + 1.0)*(t*b3x + a1x*(-t + 1.0))), 2.0) + n*pow(d1y - t*(t*(t*d2y + c3y*(-t + 1.0)) + (-t + 1.0)*(t*c3y + b3y*(-t + 1.0))) - (-t + 1.0)*(t*(t*c3y + b3y*(-t + 1.0)) + (-t + 1.0)*(t*b3y + a1y*(-t + 1.0))), 2.0);
}

double Bezier::GetXt(double t) const
{
	assert(t <= 1.01);
	assert(t >= -0.01);
	return x[0]*(1.0-t)*(1.0-t)*(1.0-t) + x[1]*3.0*(1.0-t)*(1.0-t)*t + x[2]*3.0*(1.0-t)*t*t + x[3]*t*t*t;
	// not fast, but readable
}

double Bezier::GetYt(double t) const
{
	assert(t <= 1.01);
	assert(t >= -0.01);
	return y[0]*(1.0-t)*(1.0-t)*(1.0-t) + y[1]*3.0*(1.0-t)*(1.0-t)*t + y[2]*3.0*(1.0-t)*t*t + y[3]*t*t*t;
	// not fast, but readable
}

double Bezier::GetInvDt(double t) const
{
	assert(t <= 1.01);
	assert(t >= -0.01);
	return invD[0]*(1.0-t)*(1.0-t)*(1.0-t) + invD[1]*3.0*(1.0-t)*(1.0-t)*t + invD[2]*3.0*(1.0-t)*t*t + invD[3]*t*t*t;
	// not fast, but readable
}

double Bezier::GetDerivt(double t) const
{
	assert(t <= 1.01);
	assert(t >= -0.01);
	double dx = x[0]*(-3.0)*(1.0-t)*(1.0-t) + x[1]*3.0*(1.0-4.0*t+3.0*t*t) + x[2]*3.0*t*(2.0-3.0*t) + x[3]*3.0*t*t;
	double dy = y[0]*(-3.0)*(1.0-t)*(1.0-t) + y[1]*3.0*(1.0-4.0*t+3.0*t*t) + y[2]*3.0*t*(2.0-3.0*t) + y[3]*3.0*t*t;
	return sqrt(dx*dx+dy*dy);
	// not fast, but readable
}