// (c) 2002 by Jrn Loviscach

#if !defined(BITMAP_H)
#define BITMAP_H

#include <string>
#include <assert.h>
#include "gbuffer.h"

class Bitmap
{
	public:
	Bitmap();
	Bitmap(const Bitmap& in);
	~Bitmap();
	void CreateBlack (unsigned short width, unsigned short height);
	void CreateWhite (unsigned short width, unsigned short height);
	void White();
	void Frame(const Bitmap& in);
	unsigned char GetPixel(int x, int y, int channel) const;
	void SetPixel(int x, int y, int channel, unsigned char value);
	unsigned short GetWidth() const { return width_; }
	unsigned short GetHeight() const { return height_; }
	void Extract(const GBuffer& in, double depthSens, double angleSens, double minDepth, double maxDepth, bool intersctAsCreas);
	void Clean(int gapAroundSilh);
	void Thin();
	void Blend(const Bitmap& bm);
	const unsigned char* DataPointer() const { return imageData; }
	void InitFromRGBA(unsigned char* RGBA, int width, int height);
	void StoreToRGBA(unsigned char* RGBA);

	private:
	bool hasBMP;
	unsigned char* imageData;
	unsigned short width_;
	unsigned short height_;
	unsigned long rowLength;
	bool thinOnce(Bitmap* out, int channel) const;
};

inline unsigned char Bitmap::GetPixel(int x, int y, int channel) const
{
	assert(hasBMP);
	assert( (x>=0) && (x<width_) && (y>=0) && (y<height_) && (channel>=0) && (channel<3) );

	return imageData[3*x + rowLength*y + channel];
}

inline void Bitmap::SetPixel(int x, int y, int channel, unsigned char value)
{
	assert(hasBMP);
	assert( (x>=0) && (x<width_) && (y>=0) && (y<height_) && (channel>=0) && (channel<3) );

	imageData[3*x + rowLength*y + channel]= value;
}

#endif