// (c) 2002 by Jrn Loviscach

#if !defined(BRUSH_H)
#define BRUSH_H

#include <string>
class Bitmap;
class BezierPathCollection;

const int numberOfBrushTypes = 4;
class Brush;
extern Brush* GetBrush(int i);

class Brush // abstract class
{
	public:
	Brush() {};
	virtual ~Brush();
	virtual std::string GetName() const = 0;
	union Param {float f_; int i_;};
	virtual int GetNumParams() const;
	virtual std::string GetParamName(int i) const;
	virtual bool IsIntParam(int i) const;
	virtual double MinParamVal(int i) const;
	virtual double MaxParamVal(int i) const;
	virtual double StdParamVal(int i) const;
	virtual Param* GetCopyOfStdParams() const;
	virtual double* ConvParamsIFtoD(const Param* params) const;	
	virtual void Paint(
		const BezierPathCollection& bezColl,
		Bitmap& bm,
		const double* params,
		double time = 0.0,
		double targetDpi = 72,
		double gamma = 2.2) const = 0;

	protected:
	virtual void clampParams(double* params) const;
	struct ParamData {std::string* name; bool isInt; double minVal; double maxVal; double stdVal;};
	int numParams;
	ParamData* myParams;
};

class Pen : public Brush
{
	public:
	Pen();
	virtual ~Pen() {};
	virtual std::string GetName() const;
	virtual void Paint(
		const BezierPathCollection& bezColl,
		Bitmap& bm,
		const double* params,
		double time = 0.0,
		double targetDpi = 72,
		double gamma = 2.2) const;
};

class InkBrush : public Brush
{
	public:
	InkBrush();
	virtual ~InkBrush() {};
	virtual std::string GetName() const;
	virtual void Paint(
		const BezierPathCollection& bezColl,
		Bitmap& bm,
		const double* params,
		double time = 0.0,
		double targetDpi = 72,
		double gamma = 2.2) const;
};

class WiggleLines : public Brush
{
	public:
	WiggleLines();
	virtual ~WiggleLines() {};
	virtual std::string GetName() const;
	virtual void Paint(
		const BezierPathCollection& bezColl,
		Bitmap& bm,
		const double* params,
		double time = 0.0,
		double targetDpi = 72,
		double gamma = 2.2) const;
};

class Pencil : public Brush
{
	public:
	Pencil();
	virtual ~Pencil() {};
	virtual std::string GetName() const;
	virtual void Paint(
		const BezierPathCollection& bezColl,
		Bitmap& bm,
		const double* params,
		double time = 0.0,
		double targetDpi = 72,
		double gamma = 2.2) const;
};

#endif