// (c) 2002 by Jrn Loviscach

#include <cstring> // for memset
#include <exception>
#include "assert.h"

#include "gbuffer.h"
#include "adapt.h"
#include "bitmap.h"
#include "pixelpathcollection.h"
#include "bezierpathcollection.h"
#include "brush.h"

GBuffer::GBuffer(long width, long height) // throws exceptions
: w_(width), h_(height)
{
	// to improve: if one fails, the memory allocted before is lost

	RGBA = BNEW(unsigned char, 4*w_*h_);
	depth = BNEW(float, w_*h_);
	objNumber = BNEW(long, w_*h_);
	normalX = BNEW(float, w_*h_);
	normalY = BNEW(float, w_*h_);
	normalZ = BNEW(float, w_*h_);
}

GBuffer::~GBuffer()
{
	BDELETE(normalZ);
	BDELETE(normalY);
	BDELETE(normalX);
	BDELETE(objNumber);
	BDELETE(depth);
	BDELETE(RGBA);
}

void GBuffer::Clear()
{
	memset(objNumber, 0xFF, GetWidth()*GetHeight()*sizeof(long));
	// object number 0xFFFFFFFF = -1 means: no object here

	// note: RGBA buffer is initialized by reading the rendered picture
}

// just return the address of our row
// can be used to read and write
unsigned char* GBuffer::RenderedRow(long y)
{
	assert(y>=0 && y<GetHeight());
	return RGBA+4*GetWidth()*y;
}

// returns 0 if OK
long GBuffer::Render(const Params* myParams, float time, void (*setProgressBar)(long))
{
	try
	{
		setProgressBar(0);
		Bitmap compPic;
		////////AUS RENDER NEHMEN FR COMPOSITE!!!!
		//compositePic.InitFromRGBA(RGBA, GetWidth(), GetHeight());
		compPic.CreateWhite(GetWidth(), GetHeight());

		Bitmap extractPic;
		extractPic.CreateBlack(GetWidth()+2, GetHeight()+2);
		extractPic.Extract(*this, myParams->depthDiff, myParams->angleDiff, myParams->minDepth, myParams->maxDepth, myParams->intersctAsCreas != 0);
		setProgressBar(10);
		extractPic.Clean(myParams->silhGap);
		extractPic.Thin();
		setProgressBar(20);

		PixelPathCollection silhPixels;
		silhPixels.Trace(extractPic, 0, myParams->depthDiff, myParams->minLength);
		setProgressBar(30);
		
		BezierPathCollection silhBeziers;
		silhBeziers.Trace(silhPixels, myParams->fitTolerance);
		setProgressBar(40);

		PixelPathCollection creasePixels;
		creasePixels.Trace(extractPic, 1, myParams->depthDiff, myParams->minLength);
		setProgressBar(50);
		
		BezierPathCollection creaseBeziers;
		creaseBeziers.Trace(creasePixels, myParams->fitTolerance);
		setProgressBar(55);

		double* params = GetBrush(myParams->silhBrush)->ConvParamsIFtoD(myParams->silhBrushParams);
		GetBrush(myParams->silhBrush)->Paint(silhBeziers, compPic, params, time);
		BDELETE(params);
		setProgressBar(60);

		params = GetBrush(myParams->creaseBrush)->ConvParamsIFtoD(myParams->creaseBrushParams);
		GetBrush(myParams->creaseBrush)->Paint(creaseBeziers, compPic, params, time);
		BDELETE(params);
		setProgressBar(80);

		setProgressBar(100);
		compPic.StoreToRGBA(RGBA);
	}
	catch(exception&) // most probably out of memory
	{
		return -1;
	}

	return 0;
}