// (c) 2002 by Jrn Loviscach

#include "adapt.h"
#include "pixelpathcollection.h"
#include "pixelpath.h"
#include "bitmap.h"

PixelPathCollection::PixelPathCollection()
: hasPath(false)
{}

PixelPathCollection::~PixelPathCollection()
{
	if(hasPath)
	{
		hasPath = false;
		for(int i=0; i<pathColl.size(); ++i)
		{
			DELETE(pathColl[i]);
		}
	}
}

// Bitmap is cleared by this operation!
// "framing" is removed from result
void PixelPathCollection::Trace(const Bitmap& in, int channel, int depthDiff, int minLength)
{
	if(hasPath)
	{
		hasPath = false;
		for(int i=pathColl.size()-1; i>=0; --i)
		{
			delete pathColl[i];
			pathColl.pop_back();
		}
	}
	assert(pathColl.empty());

	Bitmap bm(in);
	
	PixelPath* path;
	while( path = NEW(PixelPath), path->Trace(bm, channel, depthDiff) )
	{
		if(path->GetLength() >= minLength)
		{
			path->Unframe();
			path->ConformDirection();
			pathColl.push_back(path);
		}
		else
		{
			delete path;
		}
	}

	width_ = bm.GetWidth()-2; // remove framing
	height_ = bm.GetHeight()-2;
	hasPath = true;
}

void PixelPathCollection::CreateTestBitmap(Bitmap& bm) const
{
	assert(hasPath);

	bm.CreateBlack(width_, height_);

	for (int j=0; j<pathColl.size(); ++j)
	{
		for(int i=0; i<pathColl[j]->GetLength(); ++i)
		{
			bm.SetPixel(pathColl[j]->GetX(i), pathColl[j]->GetY(i), 0, 0x80+((59*j+123)&0x7F));
			bm.SetPixel(pathColl[j]->GetX(i), pathColl[j]->GetY(i), 1, 0x80+((41*j+213)&0x7F));
			bm.SetPixel(pathColl[j]->GetX(i), pathColl[j]->GetY(i), 2, 0x80+((83*j+98)&0x7F));
		}
	}
}