// (c) 2002 Jrn Loviscach

#include "c4d_tools.h"
#include "c4d_general.h"
#include "c4d_memory.h"
#include "c4d_resource.h"
#include "c4d_symbols.h"
#include "c4d_basecontainer.h"
#include "c4d_gui.h"
#include "ge_math.h"

#include "gbuffer.h"
#include <stdexcept>

#include "params.h"
#include "brush.h"
#include "brushdialog.h"

static const int plugInID = 1009181;

enum {
	DEPTH_DIFF=100,
	ANGLE_DIFF,
	MIN_DEPTH,
	MAX_DEPTH,
	SILH_GAP,
	INTERSCT_AS_CREAS,
	MIN_LENGTH,
	FIT_TOLERANCE,
	SILH_BRUSH,
	CREASE_BRUSH,

	SILH_BRUSH_PARAMS = 1000, // a sub-container
	CREASE_BRUSH_PARAMS
};

static Bool AllocSettings(VP_DATA *s)
{
	BaseContainer settings(s->settings, TRUE); //TRUE: create link
	settings.SetReal(DEPTH_DIFF, 10.0);
	settings.SetReal(ANGLE_DIFF, 20.0*pi/180.0);
	settings.SetReal(MIN_DEPTH, 10.0);
	settings.SetReal(MAX_DEPTH, 1000.0);
	settings.SetReal(SILH_GAP, 5.0);
	settings.SetBool(INTERSCT_AS_CREAS, FALSE);
	settings.SetReal(MIN_LENGTH, 5.0);
	settings.SetReal(FIT_TOLERANCE, 1.0);
	settings.SetLong(SILH_BRUSH, 0);
	settings.SetLong(CREASE_BRUSH, 0);

	Brush::Param* stdParams = GetBrush(0)->GetCopyOfStdParams();
	BaseContainer silhBrushParams, creaseBrushParams;
	INT i;
	for( i = 0; i < GetBrush(0)->GetNumParams(); ++i )
	{
		if(GetBrush(0)->IsIntParam(i))
		{
			silhBrushParams.SetLong(1000+i, stdParams[i].i_);
			creaseBrushParams.SetLong(1000+i, stdParams[i].i_);
		}
		else
		{
			silhBrushParams.SetReal(2000+i, stdParams[i].f_);
			creaseBrushParams.SetReal(2000+i, stdParams[i].f_);
		}
	}
	bDelete(stdParams);
	settings.SetContainer(SILH_BRUSH_PARAMS, silhBrushParams);
	settings.SetContainer(CREASE_BRUSH_PARAMS, creaseBrushParams);

	return TRUE;
}

static Params* ParamsFromContainer(VP_DATA *s, Params* myParams)
{
	BaseContainer settings(s->settings, TRUE); //TRUE: create link

	myParams->depthDiff = settings.GetReal(DEPTH_DIFF);
	myParams->angleDiff = settings.GetReal(ANGLE_DIFF);
	myParams->minDepth = settings.GetReal(MIN_DEPTH);
	myParams->maxDepth = settings.GetReal(MAX_DEPTH);
	myParams->silhGap = settings.GetReal(SILH_GAP);
	myParams->intersctAsCreas = settings.GetBool(INTERSCT_AS_CREAS);
	myParams->minLength = settings.GetLong(MIN_LENGTH);
	myParams->fitTolerance = settings.GetReal(FIT_TOLERANCE);
	myParams->silhBrush = settings.GetLong(SILH_BRUSH);
	myParams->creaseBrush = settings.GetLong(CREASE_BRUSH);

	bDelete(myParams->silhBrushParams);
	myParams->silhBrushParams
		= bNew Brush::Param[GetBrush(myParams->silhBrush)->GetNumParams()];
	bDelete(myParams->creaseBrushParams);
	myParams->creaseBrushParams
		= bNew Brush::Param[GetBrush(myParams->creaseBrush)->GetNumParams()];
	INT i;
	for(i = 0; i < GetBrush(myParams->silhBrush)->GetNumParams(); ++i)
	{
		if(GetBrush(myParams->silhBrush)->IsIntParam(i))
		{
			myParams->silhBrushParams[i].i_ =
				settings.GetContainer(SILH_BRUSH_PARAMS).GetLong(1000+i);
		}
		else
		{
			myParams->silhBrushParams[i].f_ =
				settings.GetContainer(SILH_BRUSH_PARAMS).GetReal(2000+i);
		}
	}
	for(i = 0; i < GetBrush(myParams->creaseBrush)->GetNumParams(); ++i)
	{
		if(GetBrush(myParams->creaseBrush)->IsIntParam(i))
		{
			myParams->creaseBrushParams[i].i_ =
				settings.GetContainer(CREASE_BRUSH_PARAMS).GetLong(1000+i);
		}
		else
		{
			myParams->creaseBrushParams[i].f_ =
				settings.GetContainer(CREASE_BRUSH_PARAMS).GetReal(2000+i);
		}
	}

	return myParams;
}

static void SetProgressBar(LONG p)
{
	StatusSetBar(p);
	StatusSetText("12point0 Postprocessing"); /////KOMMT NICHT IMMER AN?!?!?
}

static LONG Execute(VP_DATA *s, VideoPostStruct *vps)
{
	////////// vps->thread->TestBreak() UM AUF ABBRUCH ZU TESTEN

	switch(vps->vp)
	{
	case VP_FRAMESEQUENCE:
		if (vps->open)
		{			
			LONG width = vps->render->GetBuffer(VPBUFFER_RGBA, 0)->GetBw();
			// for whatever reason, the height is returned too low by 1
			LONG height = 1 + vps->render->GetBuffer(VPBUFFER_RGBA, 0)->GetBh();

			try
			{
				// remember our buffer in VP_DATA
				s->render = gNew GBuffer(width, height);
			}
			catch(exception&) // internal exception in GBuffer
			{
				return RAY_NOMEM; // most probably because of low memory
			}

			if(s->render == NULL)
			{
				return RAY_NOMEM;
			}
		}
		else
		{
			GBuffer* gBuf = (GBuffer*) (s->render);
			gDelete(gBuf);
			gBuf = NULL;
		}
		break;
	case VP_FRAME:
		if ( !vps->open ) // closing a frame: calculate and composite cartoon effect
		{
			VPBuffer* rgbaBuf = vps->render->GetBuffer(VPBUFFER_RGBA, 0);
			GBuffer* gBuf = (GBuffer*) (s->render);

			Params* myParams = gNew Params();
			if(myParams == 0)
			{
				return RAY_NOMEM;
			}

			LONG y = -1;
			// get rendered picture
			for(y = 0; y < gBuf->GetHeight(); ++y)
			{
				rgbaBuf->GetLine(0, y, gBuf->GetWidth(),
					(void*)(gBuf->RenderedRow(y)), 8, TRUE);
			}

			if(gBuf->Render(ParamsFromContainer(s, myParams), vps->time.Get(), SetProgressBar))
			{
				return RAY_NOMEM;
			}
			gDelete(myParams);

			// write composited picture
			for(y = 0; y < gBuf->GetHeight(); ++y)
			{
				rgbaBuf->SetLine(0, y, gBuf->GetWidth(),
					(void*)(gBuf->RenderedRow(y)), 8, TRUE);
			}
			
			// ugly, ugly
			if(vps->editor_render)
			{
				MessageDialog("12point0:\n To see the result, please minimize\n and maximize the window of the application.\n This is due to a bad redraw\n implementation in Cinema.");
			}
		}
		break;
	case VP_SUBFRAME:
		// only use last subframe for cartoon (looks better with motion blur)
		if ( vps->open && (vps->subframe == vps->subframe_cnt-1) )
		{
			// we're at the start of the last subframe: prepare to render
			((GBuffer*) (s->render))->Clear();
		}
		break;
	case VP_RENDER:
		break;
	case VP_INNER:
		break;
	}
	
	return RAY_OK;	
}

static void ExecutePixel(VP_DATA* s, PixelPost* pp, LONG x, LONG subx, LONG suby)
{
	GBuffer* gBuf = (GBuffer*)(s->render);
	long y = pp->line;

	if(*(pp->frag) == NULL) // no object found under the ray
	{
		gBuf->SetObjNumber(x, y, -1);
		return;
	}

	VolumeData* myVd = pp->vd;

	LONG vind = 0;
	gBuf->SetObjNumber(x, y,
		myVd->Obj_to_Num(myVd->ID_to_Obj(myVd->lhit, &vind)));

	gBuf->SetDepth(x, y, myVd->dist);
	gBuf->SetNormalX(x, y, myVd->n.x);
	gBuf->SetNormalY(x, y, myVd->n.y);
	gBuf->SetNormalZ(x, y, myVd->n.z);
}

class VideoPostDialog: public SubDialog
{
	public:
		BaseContainer data;

		VideoPostDialog(void) {}
		virtual ~VideoPostDialog(void) {}

		void SetData(_BaseContainer *t_data) { data = BaseContainer(t_data, TRUE); }

		virtual Bool CreateLayout(void);
		virtual Bool InitValues(void);
		virtual Bool Command(LONG id,const BaseContainer &msg);
};

Bool VideoPostDialog::CreateLayout(void)
{
	Bool isOK = LoadDialogResource(DLG_12POINT0, NULL, BFH_SCALEFIT | BFV_SCALEFIT);
	INT i;
	for(i=0; i<numberOfBrushTypes; ++i)
	{
		Brush* b = GetBrush(i);
		if(b == NULL)
		{
			return 0;
		}
		String name = String((b->GetName()).c_str());
		isOK = isOK && AddChild(IDC_SILH_BRUSH, i, name)
			&& AddChild(IDC_CREASE_BRUSH, i, name);
	}
	return isOK;
}

Bool VideoPostDialog::InitValues(void)
{
	return
	   SetMeter(IDC_DEPTH_DIFF, &data, DEPTH_DIFF, 1e-10)
	&& SetDegree(IDC_ANGLE_DIFF, &data, ANGLE_DIFF, 1.0)
	&& SetMeter(IDC_MIN_DEPTH, &data, MIN_DEPTH, 1e-10)
	&& SetMeter(IDC_MAX_DEPTH, &data, MAX_DEPTH, 1e-10)
	&& SetReal(IDC_SILH_GAP, &data, SILH_GAP, 1, 10)
	&& SetBool(IDC_INTERSCT_AS_CREAS, &data, INTERSCT_AS_CREAS)
	&& SetLong(IDC_MIN_LENGTH, &data, MIN_LENGTH, 1, 50)
	&& SetReal(IDC_FIT_TOLERANCE, &data, FIT_TOLERANCE, 0.01, 1000.0)
	&& SetLong(IDC_SILH_BRUSH, &data, SILH_BRUSH)
	&& SetLong(IDC_CREASE_BRUSH, &data, CREASE_BRUSH);
}

Bool VideoPostDialog::Command(LONG id, const BaseContainer &msg)
{
	Bool isOK = TRUE;

	switch (id)
	{
		case IDC_DEPTH_DIFF:
			isOK = isOK && GetReal(id, &data, DEPTH_DIFF);
			break;
		case IDC_ANGLE_DIFF:
			isOK = isOK && GetReal(id, &data, ANGLE_DIFF);
			break;
		case IDC_MIN_DEPTH:
			isOK = isOK && GetReal(id, &data, MIN_DEPTH);
///////////NOCH VERHINDERN, DASS MIN_DEPTH >= MAX_DEPTH WIRD
			break;
		case IDC_MAX_DEPTH:
			isOK = isOK && GetReal(id, &data, MAX_DEPTH);
///////////NOCH VERHINDERN, DASS MIN_DEPTH >= MAX_DEPTH WIRD
			break;
		case IDC_SILH_GAP:
			isOK = isOK && GetReal(id, &data, SILH_GAP);
			break;
		case IDC_INTERSCT_AS_CREAS:
			isOK = isOK && GetBool(id, &data, INTERSCT_AS_CREAS);
			break;
		case IDC_MIN_LENGTH:
			isOK = isOK && GetLong(id, &data, MIN_LENGTH);
			break;
		case IDC_FIT_TOLERANCE:
			isOK = isOK && GetReal(id, &data, FIT_TOLERANCE);
			break;
		case IDC_SILH_BRUSH:
			{
				long oldBrush = data.GetLong(SILH_BRUSH);
				long newBrush = -1;
				isOK = isOK && GetLong(id, newBrush);
				if(oldBrush != newBrush)
				{
					data.SetLong(SILH_BRUSH, newBrush);
					// new brush, so load standard parameter values

					Brush::Param* stdParams = GetBrush(newBrush)->GetCopyOfStdParams();
					BaseContainer brushParams;

					INT i;
					for( i = 0; i < GetBrush(newBrush)->GetNumParams(); ++i )
					{
						if(GetBrush(newBrush)->IsIntParam(i))
						{
							brushParams.SetLong(1000+i, stdParams[i].i_);
						}
						else
						{
							brushParams.SetReal(2000+i, stdParams[i].f_);
						}
					}
					BDELETE(stdParams);
					data.SetContainer(SILH_BRUSH_PARAMS, brushParams);
				}
			}
			break;
		case IDC_CREASE_BRUSH:
			{
				long oldBrush = data.GetLong(CREASE_BRUSH);
				long newBrush = -1;
				isOK = isOK && GetLong(id, newBrush);
				if(oldBrush != newBrush)
				{
					data.SetLong(CREASE_BRUSH, newBrush);
					// new brush, so load standard parameter values

					Brush::Param* stdParams = GetBrush(newBrush)->GetCopyOfStdParams();
					BaseContainer brushParams;

					INT i;
					for( i = 0; i < GetBrush(newBrush)->GetNumParams(); ++i )
					{
						if(GetBrush(newBrush)->IsIntParam(i))
						{
							brushParams.SetLong(1000+i, stdParams[i].i_);
						}
						else
						{
							brushParams.SetReal(2000+i, stdParams[i].f_);
						}
					}
					BDELETE(stdParams);
					data.SetContainer(CREASE_BRUSH_PARAMS, brushParams);
				}
			}
			break;
		case IDC_EDIT_SILH:
			{
				BrushDialog d(data.GetLong(SILH_BRUSH),
					data.GetContainerInstance(SILH_BRUSH_PARAMS), "Silhouette");
				d.Open(FALSE, 0);
			}
			break;
		case IDC_EDIT_CREASE:
			{
				BrushDialog d(data.GetLong(CREASE_BRUSH),
					data.GetContainerInstance(CREASE_BRUSH_PARAMS), "Crease");
				d.Open(FALSE, 0);
			}
			break;
		case IDC_SILH_TO_CREASE_BR:
			{
				long newBrush = data.GetLong(SILH_BRUSH);
				data.SetLong(CREASE_BRUSH, newBrush);
				isOK = isOK && SetLong(IDC_CREASE_BRUSH, newBrush);

				BaseContainer brushParams(data.GetContainer(SILH_BRUSH_PARAMS));
				data.SetContainer(CREASE_BRUSH_PARAMS, brushParams);
			}
			break;
		case IDC_CREASE_TO_SILH_BR:
			{
				long newBrush = data.GetLong(CREASE_BRUSH);
				data.SetLong(SILH_BRUSH, newBrush);
				isOK = isOK && SetLong(IDC_SILH_BRUSH, newBrush);

				BaseContainer brushParams(data.GetContainer(CREASE_BRUSH_PARAMS));
				data.SetContainer(SILH_BRUSH_PARAMS, brushParams);
			}
			break;
	}
	return isOK;
}

static CDialog *AllocDialog(_BaseContainer *data, CDialog *cd_parent, LONG dlg_id)
{
	VideoPostDialog *dlg = gNew VideoPostDialog;
	if (!dlg) return NULL;

	dlg->SetData(data);
	C4DOS.Cd->AttachSubDialog(cd_parent, dlg_id, dlg->Get());

	return dlg->Get();
}

static void FreeDialog(CDialog *cd)
{
	VideoPostDialog *dlg = (VideoPostDialog*)C4DOS.Cd->GetUserData(cd);
	gDelete(dlg);
}


Bool RegisterVideoPost(void)
{
	String name = GeLoadString(IDS_12POINT0);
	if (!name.Content()) return TRUE;

	VIDEOPOST vp;
	ClearMem(&vp, sizeof(vp));

	vp.level            = 1;
	vp.info             = VIDEOPOST_REFRESH; // regrettably, doesn't help with VP_FRAME
	vp.Execute          = Execute;
	vp.AllocateBuffers  = NULL;
	vp.AllocSettings    = AllocSettings;
	vp.ExecutePixel     = ExecutePixel;
	vp.AllocDialog      = AllocDialog;
	vp.FreeDialog       = FreeDialog;
	vp.priority         = 0;

	MessageDialog("12point0:\n This is an early pre-release.\n Do not use. Do not distribute.\n\n Thank you.\n Jrn Loviscach\n j.loviscach@computer.org");

	return GeRegisterPlugin(C4DPL_VIDEOPOST, plugInID, name, &vp, sizeof(vp));
}
